/* ddist.cc: Compute _global_ distribution of distances. */

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include "ptopk.h"
#include "histogram.h"
#include "index.h"
#include "ddist.h"

int random_int(int min, int max)
{
     return random() % (max - min) + min;
}

static double square(double x) { return x * x; }

static inline
float point_distance(Point a, Point b)
{
     assert(a.p != -1 && b.p != -1);
     double d = sqrt(square(a.x1 - b.x1) + square(a.x2 - b.x2));
     if (d > 1000)
          printf("(%f, %f) (%f, %f); d = %f\n", a.x1, a.x2, b.x1, b.x2, d);
     return d;
}

struct histogram *ddist(ReadableIndex *a, ReadableIndex *b, int sample_size)
{
     float *distances = new float[sample_size];
     for (int si = 0; si < sample_size; si++) {
          int ai, bi;
          for (;;) {
               ai = random_int(0, a->npoints() - 1);
               bi = random_int(0, b->npoints() - 1);
               if (a->get_point(ai).p != -1 && b->get_point(bi).p != -1)
                    break;
          }
//          a->get_point(ai).print(); printf("\n");
          distances[si] = point_distance(a->get_point(ai), b->get_point(bi));
//          printf("%d %d %f\n", ai, bi, distances[si]);
     }
     struct histogram *h = histogram(distances, sample_size);
//     plot_pdf(h);
     delete distances;
     return h;
}

DistDistHistogram::DistDistHistogram(const std::string astem, 
                                     const std::string bstem, 
                                     int sample_size) {
     ReadableIndex a(astem), b(bstem);
     h = ddist(&a, &b, sample_size);
}

DistDistHistogram::~DistDistHistogram() { 
     free(h); 
}
void DistDistHistogram::plot() { 
     plot_pdf(h); 
}

double DistDistHistogram::score_ub(Box box1, Box box2)
{
     return box1.p_max * box2.p_max * 
          (1 -  pvalue(box_min_distance(box1, box2)));
}

double DistDistHistogram::score_lb(Box box1, Box box2)
{
     return box1.p_min * box2.p_min * 
          (1 -  pvalue(box_max_distance(box1, box2)));
}

float DistDistHistogram::pvalue(float distance)
{
     float pv = histogram_pvalue(h, distance);
     assert(pv >= 0.0 && pv <= 1.0);
     return pv;
}

//
//
//

DistDistExp::DistDistExp(double lambda)
     : lambda(lambda)
{
}

double DistDistExp::score_ub(Box box1, Box box2)
{
     double ub = box1.p_max * box2.p_max * 
          lambda * exp(-lambda * box_min_distance(box1, box2));
     assert(ub >= 0 && ub <= 1);
     return ub;
}

double DistDistExp::score_lb(Box box1, Box box2)
{
     double lb = box1.p_min * box2.p_min * 
          lambda * exp(-lambda * box_max_distance(box1, box2));
     assert(lb >= 0 && lb <= 1);
     return lb;
}

float DistDistExp::pvalue(float distance)
{
     double pv = 1 - lambda * exp(-lambda * distance);
     assert(pv >= 0 && pv <= 1);
//     printf("pv(%f) = %f\n", distance, pv);
     return pv;
}
