/* Perturb the horizontal cell dataset in order to increase its size
 * and density. */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <sysexits.h>
#include <vlcutils/error.h>
#include <string>
#include <iostream>
#include <getopt.h>
#include "mmapping.h"
#include "ptopk.h"
#include "index.h"

static void usage(FILE *stream)
{
     fprintf(stream ,"Usage %s [-ahv1] INPUT-STEM OUTPUT-STEM FACTOR\n", 
                  program_name);
     if (stream == stderr)
          fprintf(stream, "Try `%s -h` for more information.\n", program_name);
}

static void help(void)
{
     usage(stdout);
     printf("\n"
"      -a            Append to the target index instead of overwriting it.\n"
"      -h            Display this help.\n"
"      -v            Be verbose.\n"
"      -1            Put all points in one object and one box.\n"
          );
}

static bool one_box = false, verbose = false, append_index = false;
static std::string input_stem, output_stem;
static int factor;

static void parse_options(int argc, char *argv[])
{
     int opt;

     while ((opt = getopt(argc, argv, "ahv1")) != -1)
          switch (opt) {
          case 'a': append_index = true; break;
          case 'h': help(); exit(EX_OK);
          case 'v': verbose = true; break;
          case '1': one_box = true; break;
          default: usage(stderr); exit(EX_USAGE);
          }
     if (argc == optind + 3) {
          input_stem = argv[optind]; 
          output_stem = argv[optind + 1];
          factor = atoi(argv[optind + 2]);
          if (factor < 1)
               fatal_error("Factor must be at least one: %s",
                           argv[optind + 2]);
     } else {
          usage(stderr);
          exit(EX_USAGE);
     }
         
}

extern "C" { double ltqnorm(double); }

#define T_MEAN 50
#define T_STDDEV 10

int main(int argc, char *argv[])
{
     set_program_name(argv[0]);
     parse_options(argc, argv);

     ReadableIndex in(input_stem);
     if (verbose) in.info();

     // Block to ensure the index is closed before we print its info.
     {
          WritableIndex out(output_stem,
                            one_box ? in.npoints() * factor : in.box_size,
                            append_index ? append : supersede);
          srandom(time(NULL));

          for (int i = 0; i < factor; i++) {
               int lbox = 0;
               for (int oid = 0; oid < in.nobjects; oid++) {
                    // Decide on a shift for this object.
                    double translation = ltqnorm(1.0 * random() / RAND_MAX) * T_STDDEV + T_MEAN;
                    double angle = 2 * M_PI * random() / RAND_MAX;
                    double delta_x = translation * cos(angle);
                    double delta_y = translation * sin(angle);

                    int k;
                    for (k = 0; k < in.nboxes(oid); k++) {
                         Box box = in.get_box(lbox + k);
                         for (int l = 0; l < box.npoints; l++) {
                              Point p = box.get_point(l);
                              Point q(p.p, p.x1 + delta_x, p.x2 + delta_y);
                              out.addPoint(q);
                         }
                    }
                    if (!one_box)
                         out.commitObject();
                    lbox += k;
               }
          }
          if (one_box)
               out.commitObject();
     }
     
     ReadableIndex out(output_stem);
     if (verbose) out.info();
     
     return EX_OK;
}

