#include <assert.h>
#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <vlcutils/error.h>
#include <vlcutils/io.h>
#include <vlcutils/mem.h>
#include <string>
#include <iostream>
#include "mmapping.h"
#include "ptopk.h"
#include "index.h"

// Note that struct point and struct box (from ptop.h) are the
// external representations of points and boxes, respectively.  The
// classes in this file serve as internal representations only.

Point nullPoint = Point(-1, -1, -1);

static int count_points(const struct point *points, int box_size)
{
     int i;
     for (i = box_size - 1; i >= 0; i--)
          if (points[i].p != -1)
               break;
     return i + 1;
}

Box::Box(const BoxSource *index, int page_id, const struct box *box_data, 
         const struct point *points)
     : p_min(box_data->p_min), p_max(box_data->p_max),
       x1_min(box_data->x1_min), x1_max(box_data->x1_max),
       x2_min(box_data->x2_min), x2_max(box_data->x2_max),
       npoints(count_points(points, index->box_size)),
       points(points), index(index), page_id(page_id),
       my_object_id(-1), box_size(index->box_size)
     
{
     assert(box_data && points);
}

Box::Box(int box_size, int object_id, const struct point *points)
     : npoints(count_points(points, box_size)),
       index(NULL), page_id(-1), my_object_id(object_id), box_size(box_size),
       points(points)
{
     for (int i = 0; i < npoints; i++) {
          if (i == 0 || points[i].p < p_min) p_min = points[i].p;
          if (i == 0 || points[i].p > p_max) p_max = points[i].p;
          if (i == 0 || points[i].x1 < x1_min) x1_min = points[i].x1;
          if (i == 0 || points[i].x1 > x1_max) x1_max = points[i].x1;
          if (i == 0 || points[i].x2 < x2_min) x2_min = points[i].x2;
          if (i == 0 || points[i].x2 > x2_max) x2_max = points[i].x2;
     }
}

Point Box::get_point(int i)
{
     return Point(&points[i]);
}

int Box::object_id() const
{
     return my_object_id == -1 ? index->object_id(page_id) : my_object_id;
}

double Box::volume() const
{
     return (p_max - p_min) * (x1_max - x1_min) * (x2_max - x2_min);
}

Point::Point() : p(-1), x1(-1), x2(-1) { }

Point::Point(float p, float x1, float x2) : p(p), x1(x1), x2(x2) {}

Point::Point(const struct point *point)
     : p(point->p), x1(point->x1), x2(point->x2)
{}

// bool Point::operator==(const Point &other) const
// {
//      return p == other.p && x1 == other.x1 && x2 == other.x2;
// }

bool Point::operator!=(const Point &other) const
{
     return p != other.p || x1 != other.x1 || x2 != other.x2;
}

void Point::write(struct point *external) {
     external->p = p;
     external->x1 = x1;
     external->x2 = x2;
}

void Point::print() {
     printf("[%f (%d, %d)]", p, x1, x2);
}

static int index_box_size(const std::string filename)
{
     Mmapping<struct box> boxes(filename.c_str(), "/boxes");
     Mmapping<struct point> points(filename.c_str(), "/points");
     int box_size = points.size / boxes.size;
     return box_size;
}

ReadableIndex::ReadableIndex(const std::string filename) 
     : objects(new Mmapping<int>(filename.c_str(), "/objects")),
       nobjects(objects->size),
       boxes(new Mmapping<struct box>(filename.c_str(), "/boxes")),
       points(new Mmapping<struct point>(filename.c_str(), "/points")),
       npoints_cached(-1), points_per_box(NULL), filename(filename),
       BoxSource(index_box_size(filename))
{
     assert(points->size % boxes->size == 0);
}

ReadableIndex::~ReadableIndex()
{
     if (points_per_box)
          delete points_per_box;
     delete objects;
     delete boxes;
     delete points;
}

void ReadableIndex::info() {
     std::cout << filename << ": " << nobjects << " object(s), "
               << nboxes() << " box(es), "
               << npoints() << " point(s).  "
               << "Box size: " << box_size << std::endl;
}

Box ReadableIndex::get_box(int page_id) {
     return Box(this, page_id, &boxes->data[page_id], 
                &points->data[page_id * box_size]);
}

int ReadableIndex::npoints() {
     if (npoints_cached == -1) {
          npoints_cached = 0;
          for (int i = 0; i < nboxes(); i++)
               npoints_cached += get_box(i).npoints;
     }
     return npoints_cached;
}

Point ReadableIndex::get_point(int point) {
     if (!points_per_box) {
          points_per_box = new int[nboxes()];
          for (int box = 0; box < nboxes(); box++)
               points_per_box[box] = get_box(box).npoints;
     }
     assert(point >= 0 && point < npoints());
     int lpoint = 0; // No. of linear points seen before this box.
     for (int i = 0; i < nboxes(); i++) {
          Box box = get_box(i);
          if (lpoint + box.npoints > point)
               return box.get_point(point - lpoint);
          lpoint += box.npoints;
     }
     assert(false);
}

int ReadableIndex::object_id(int page_id) const {
     int l = 0;
     for (int i = 0; i < objects->size; i++) {
          if (l + objects->data[i] > page_id)
               return i;
          l += objects->data[i];
     }
     abort();
}

int ReadableIndex::nboxes() const { return boxes->size; }

int ReadableIndex::nboxes(int oid) const {
     assert(oid >= 0 && oid < nobjects);
     return objects->data[oid];
}

double ReadableIndex::average_volume() {
     int n = nboxes();
     double total = 0.0;
     for (int i = 0; i < n; i++) {
          Box box = get_box(i);
          total += box.volume();
     }
     return total / (n * 1.0);
}

// Creates a new index.
WritableIndex::WritableIndex(std::string filename, int box_size, 
                             ifexists_t ifexists) {
     const char *mode = "w";
     if (mkdir(filename.c_str(), 0755)) {
          if (errno != EEXIST || ifexists == fail)
               fatal_perror("%s: mkdir", filename.c_str());
          if (ifexists == append) {
               mode = "a";
               // Check box size.
               int nboxes = Mmapping<struct box>(filename.c_str(),
                                                 "/boxes").size;
               int npoints = Mmapping<struct point>(filename.c_str(),
                                                    "/points").size;
               if (npoints != nboxes * box_size)
                    fatal_error("Cannot append using different box size.");
          }
     }
     of = fopenck((filename + "/objects").c_str(), mode);
     bf = fopenck((filename + "/boxes").c_str(), mode);
     pf = fopenck((filename + "/points").c_str(), mode);
     nobjects = 0;
     nboxes = 0;
     npoints = 0;
     this->box_size = box_size;
     points = new struct point[box_size];
}

WritableIndex::~WritableIndex() {
     delete points;
     // Don't commit here.  You'd better call commitObject(),
     // otherwise your last object may be lost.
     fclose(pf);
     fclose(bf);
     fclose(of);
}

void WritableIndex::addPoint(Point p) {
     if (npoints == box_size)
          commitBox();
     if (npoints == 0) {
          box.p_min = box.p_max = p.p;
          box.x1_min = box.x1_max = p.x1;
          box.x2_min = box.x2_max = p.x2;
     } else {
          if (p.p < box.p_min) box.p_min = p.p;
          if (p.p > box.p_max) box.p_max = p.p;
          if (p.x1 < box.x1_min) box.x1_min = p.x1;
          if (p.x1 > box.x1_max) box.x1_max = p.x1;
          if (p.x2 < box.x2_min) box.x2_min = p.x2;
          if (p.x2 > box.x2_max) box.x2_max = p.x2;
     }
     p.write(points + npoints);
     npoints++;
}

void WritableIndex::commitBox() {
     while (npoints < box_size) {
          nullPoint.write(points + npoints);
          npoints++;
     }
     // Write the current box and its points to the disk.
     fwriteck(&box, sizeof(struct box), bf);
     fwriteck(points, sizeof(struct point) * box_size, pf);
     nboxes++;
     npoints = 0;
}

void WritableIndex::commitObject() {
     if (npoints)
          commitBox();
     fwriteck(&nboxes, sizeof(int), of);
     nboxes = 0;
     nobjects++;
}

