#include <float.h>
#include "ptopk.h"
#include "join.h"

Result::Result(int object_id_a, Point a, int object_id_b, Point b, float score)
     : object_id_a(object_id_a), object_id_b(object_id_b), a(a), b(b), 
       score(score)
{}

inline
bool Result::operator!=(const Result& other) const
{
     return !(object_id_a == other.object_id_a && a == other.a &&
              object_id_b == other.object_id_b && b == other.b);
}

inline
bool same_object(Result *r1, Result *r2)
{
     return r1->object_id_a == r2->object_id_a &&
          r1->object_id_b == r2->object_id_b;
}

void Result::print()
{
     printf("%d:(%.2f,%.2f,%.2f) %d:(%.2f,%.2f,%.2f) %f\n",
            object_id_a, a.p, a.x1, a.x2,
            object_id_b, b.p, b.x1, b.x2, score);
}

// ---------------------------------------------------------------------

TopKResults::TopKResults(int k)
     : Results(FLT_MIN), iterations(0), k(k), verbose(false)
{
}

bool TopKResults::has_threshold()
{
     return threshold > FLT_MIN;
}

bool TopKResults::insert(Result result)
{
     if (verbose && iterations % 1000000 == 0) report_stats();
     iterations++;
//     if (result.score < 1e-6)
//          fprintf(stderr, "Score is %f\n", result.score);
     std::list<Result>::iterator i = results.begin();
     while (i != results.end() && result.score < i->score) {
          if (same_object(&result, &(*i))) {
//               printf("Better match already found for this object.\n");
               return false; // Better match already found for this object.
          }
          i++;
     }
     results.insert(i, result);
     // Next, remove other matches between the same objects.
     i = results.begin();
     int position = 0;
     while (i != results.end() && position < k) {
          if ((*i) != result && same_object(&result, &(*i)))
               i = results.erase(i);
          else {
               i++;
               position++;
          }
     }
     // Remove everything after the k-th element.
     results.erase(i, results.end());
//     printf("%d results. ", results.size());
     // Finally, set threshold.
     if (position == k) {
          i = results.end();
          i--;
          threshold = i->score;
//          printf("Inserted result with score %f; new threshold %f.\n", 
//                 result.score, threshold);
     } else {
          threshold = FLT_MIN;
//         printf("Inserted result with score %f; no threshold yet.\n", result.score);
     }
     return true;
}

void TopKResults::print()
{
     int j = 1;
     for (std::list<Result>::iterator i = results.begin(); 
          i != results.end(); i++) {
          printf("%2i ", j);
          i->print();
          j++;
     }
}

int TopKResults::size()
{
     return results.size();
}

void TopKResults::report_stats()
{
     printf("%d %f\n", iterations, threshold);
}

// ---------------------------------------------------------------------

TopKResultsNoObjects::TopKResultsNoObjects(int k)
     : TopKResults(k)
{
}

bool TopKResultsNoObjects::insert(Result result)
{
     if (verbose && iterations % 1000000 == 0) report_stats();
     iterations++;
     std::list<Result>::iterator i = results.begin();
     while (i != results.end() && result.score < i->score)
          i++;
     results.insert(i, result);
     // Next, remove other matches between the same objects.
     i = results.begin();
     int position = 0;
     while (i != results.end() && position < k) {
          i++;
          position++;
     }
     // Remove everything after the k-th element.
     results.erase(i, results.end());
     // Finally, set threshold.
     if (position == k) {
          i = results.end();
          i--;
          threshold = i->score;
     } else {
          threshold = FLT_MIN;
     }
     return true;
}
