#include <math.h>
#include <float.h>
#include <assert.h>
#include <algorithm>
#include <set>
#include <typeinfo>
#include <vlcutils/error.h>
#include <vlcutils/progress.h>
#include "ptopk.h"
#include "join.h"
#include "score.h"

float triangle_start(Point a, double lambda, float threshold)
{
     return (-logf(a.p) - log(lambda) + lambda * a.x1 + logf(threshold)) / lambda;
}

float triangle_end(Point a, float lambda, float threshold)
{
     return (logf(a.p) + log(lambda) + lambda * a.x1 - logf(threshold)) / lambda;
}

float triangle_mid_lnp(Point a, float lambda, float threshold)
{
     return logf(threshold) - logf(a.p) - logf(lambda);
}

void plot_plane(Box a, Box b, double lambda, float threshold)
{
     int na = a.npoints;
     for (int i = 0; i < na; i++) {
          Point p = a.get_point(i);
          printf("%f 0\n", triangle_start(p, lambda, threshold));
          printf("%f %f\n", p.x1, triangle_mid_lnp(p, lambda, threshold));
          printf("%f 0\n", triangle_end(p, lambda, threshold));
          printf("\n");
     }
     int nb = b.npoints;
     for (int i = 0; i < nb; i++) {
          Point q = b.get_point(i);
          printf("%f %f\n\n", q.x1, logf(q.p));
     }
}


class Event {
public:
     float x;
     int point_id;
     Event() {}
     Event(float x, int point_id) : x(x), point_id(point_id) {
          assert(!isnan(x));
          if (isnan(x))
               abort_error("Event coordinate is NaN.");
     }
     bool operator<(const Event& other) const { return x < other.x; }
     virtual void foo() {} // Need a virtual method to use dynamic_cast.
     void print() {
          printf("%s %f\n", typeid(*this).name(), x);
     }
};

class StartEvent : public Event {
public:
     StartEvent() {}
     StartEvent(int point_id, Point p, double lambda, double threshold)
          : Event(triangle_start(p, lambda, threshold), point_id) {}
};

class EndEvent;

class MidEvent : public Event {
public:
     float x_end; // Triangle end point.
     StartEvent start;
     MidEvent() {}
     MidEvent(int point_id, Point p, double lambda, double threshold, 
              StartEvent start)
          : Event(p.x1, point_id), start(start), 
            x_end(triangle_end(p, lambda, threshold)) {}
};

class EndEvent : public Event {
public:
     MidEvent mid;
     EndEvent() {}
     EndEvent(int point_id, Point p, double lambda, double threshold, 
              MidEvent mid)
          : Event(triangle_end(p, lambda, threshold), point_id), mid(mid) {
          assert(x == mid.x_end);
     }
};

class PointEvent : public Event {
public:
     double p;
     PointEvent() {}
     PointEvent(int point_id, Point q) : Event(q.x1, point_id), p(q.p) {}
};

class EventQueue {
protected:
     Event nullEvent;
     StartEvent *starts;
     EndEvent *ends;
     MidEvent *mids;
     PointEvent *points;
     int na, nb;
     int next_start, next_end, next_mid, next_point;
public:
     EventQueue(Box a, Box b, double lambda, double threshold);
     ~EventQueue();
     int size();
     bool empty();
     Event& pop();
};

EventQueue::EventQueue(Box a, Box b, double lambda, double threshold)
{
     na = a.npoints;
     nb = b.npoints;
     starts = new StartEvent[na];
     mids = new MidEvent[na];
     ends = new EndEvent[na];
     points = new PointEvent[nb];
     int j = 0;
     for (int i = 0; i < a.npoints; i++) { // Not same as na.
          Point p = a.get_point(i);
          starts[j] = StartEvent(i, p, lambda, threshold);
          mids[j] = MidEvent(i, p, lambda, threshold, starts[j]);
          ends[j] = EndEvent(i, p, lambda, threshold, mids[j]);
          if (starts[j].x > mids[j].x) {
               // No solution possible.
               na--;
               continue;
          }
          assert(starts[j].x <= mids[j].x && mids[j].x <= ends[j].x);
          j++;
     }
     for (int i = 0; i < nb; i++) {
          points[i] = PointEvent(i, b.get_point(i));
     }
     std::sort(starts, starts + na);
     std::sort(mids, mids + na);
     std::sort(ends, ends + na);
     std::sort(points, points + nb);
     next_start = next_end = next_mid = next_point = 0;
 }

EventQueue::~EventQueue()
{
     delete starts;
     delete mids;
     delete ends;
     delete points;
}

inline int EventQueue::size()
{
     return (na - next_start) + (na - next_mid) + (na - next_end) + 
          (nb - next_point);
}

bool EventQueue::empty()
{
     return size() == 0;
}

Event& EventQueue::pop()
{
     if (size() == 0)
          return nullEvent;

     float min_x = FLT_MAX;
     enum { q_none, q_start, q_mid, q_end, q_point } min_q = q_none;
     if (next_start < na)
          min_x = starts[next_start].x, min_q = q_start;
     if (next_mid < na && (min_q == q_none || mids[next_mid].x < min_x))
          min_x = mids[next_mid].x, min_q = q_mid;
     if (next_end < na && (min_q == q_none || ends[next_end].x < min_x))
          min_x = ends[next_end].x, min_q = q_end;
     if (next_point < nb && (min_q == q_none || points[next_point].x < min_x))
          min_x = points[next_point].x, min_q = q_point;

     switch (min_q) {
     case q_start: return starts[next_start++];
     case q_mid: return mids[next_mid++];
     case q_end: return ends[next_end++];
     case q_point: return points[next_point++];
     default: abort();
     }
}

class Slope {
public:
     float x; // Intersection with log p = 0.
     int point_id;
     Slope(float x, int point_id) : x(x), point_id(point_id) {}
     bool operator<(const Slope& other) const { 
          return x < other.x || x == other.x && point_id < other.point_id;
     }
};

class Downslope : public Slope {
public:
     Downslope(float x) : Slope(x, -1) {}
     Downslope(StartEvent& start) : Slope(start.x, start.point_id) {}
     bool operator<(const Downslope& other) const { 
          return x < other.x || x == other.x && point_id < other.point_id;
     }
     bool operator==(const Downslope& other) const {
          return point_id == other.point_id;
     }
};

class Upslope : public Slope {
public:
     Upslope(float x) : Slope(x, -1) {}
     Upslope(MidEvent& mid) : Slope(mid.x_end, mid.point_id) {}
     bool operator<(const Upslope& other) const { 
          return x < other.x || x == other.x && point_id < other.point_id;
     }
     bool operator==(const Downslope& other) const {
          return point_id == other.point_id;
     }
};

static struct timeval start_time;
static bool verbose = false;

void planesweep_verbose_on()
{
     mark_start_time(&start_time);
     verbose = true;
}

// The plane sweep solves either a threshold probabilistic spatial
// join or a top-k probabilistic spatial join, depeneding on whether
// results->threshold changes as new results are reported.
void planesweep(Box a, Box b, double lambda, Results *results)
{
     std::set<Downslope> downslopes;
     std::set<Upslope> upslopes;
     int count = 0;
     Score_1D score(lambda);
     double t0 = results->threshold;
     struct timeval time;

     int object_id_a = a.object_id();
     int object_id_b = b.object_id();
     EventQueue queue(a, b, lambda, results->threshold);
     while (queue.size() > 0) {
          Event& e = queue.pop();
          if (typeid(e) == typeid(StartEvent)) {
               StartEvent& start = dynamic_cast<StartEvent&>(e);
               downslopes.insert(Downslope(start));
          } else if (typeid(e) == typeid(MidEvent)) {
               MidEvent& mid = dynamic_cast<MidEvent&>(e);
               if (downslopes.erase(Downslope(mid.start)) == 0)
                    abort_error("Failed to erase start(%f, %d).",
                                mid.start.x, mid.start.point_id);
               upslopes.insert(Upslope(mid));
          } else if (typeid(e) == typeid(EndEvent)) {
               EndEvent& end = dynamic_cast<EndEvent&>(e);
               if (upslopes.erase(Upslope(end.mid)) == 0)
                    abort_error("Failed to erase mid.");
          } else if (typeid(e) == typeid(PointEvent)) {
               PointEvent& point = dynamic_cast<PointEvent&>(e);
               Point q = b.get_point(e.point_id);
               double qprob = q.p * t0 / results->threshold;
               double old_threshold = results->threshold;
               Downslope downx = Downslope(q.x1 + logf(qprob) / lambda);
               // If downx.x > q.x1, the threshold is too high for
               // this point to have any matches.  (Why?)
               if (downx.x <= q.x1) {
                    std::set<Downslope>::iterator ds = 
                         downslopes.lower_bound(downx);
                    while (ds != downslopes.begin()) {
                         ds--;
                         const Point& p = a.get_point(ds->point_id);
                         double s = score(p, q);
                         results->insert(Result(object_id_a, p, object_id_b, q,
                                                s));
                         count++;
                    }
               }

               Upslope upx = Upslope(q.x1 - logf(qprob) / lambda);
               // If upx.x < q.x1, the threshold is too high for this
               // point to have any matches.
               if (upx.x >= q.x1) { 
                    std::set<Upslope>::iterator us = upslopes.lower_bound(upx);
                    while (us != upslopes.end()) {
                         const Point& p = a.get_point(us->point_id);
                         double s = score(p, q);
                         results->insert(Result(object_id_a, p, object_id_b, q,
                                                s));
                         count++;
                         us++;
                    }
               }
               if (verbose && results->threshold > old_threshold) {
                    printf("%f %e\n", elapsed_time(&start_time), 
                           results->threshold);
               }
          } else abort();
     }
     if (!downslopes.empty() || !upslopes.empty())
          warn("#ds = %d, #us = %d\n", downslopes.size(), upslopes.size());
}
