#include <math.h>
#include <float.h>
#include <assert.h>
#include <sysexits.h>
#include <map>
#include <sstream>
#include <vector>
#include <iterator>
#include <algorithm>
#include <vlcutils/mem.h>
#include <vlcutils/progress.h>
#include "ptopk.h"
#include "index.h"
#include "results.h"
#include "score.h"
#include "join.h"
#include "planesweep.h"

class PM {
protected:
     ReadableIndex& a, b;
     virtual void join(int page_id_a, int page_id_b);
public:
     Score_1D score;
     int neligible;
     TopKResultsNoObjects results;
     PM(int k, double lambda, ReadableIndex &a, ReadableIndex &b);
     virtual bool joinnext();
};

class JIT_PM : public PM {
protected:
     std::vector<bool> *eligibility;
     void mark_ineligible(int page_id_a, int page_id_b);
     bool eligible(int page_id_a, int page_id_b);
     virtual void join(int page_id_a, int page_id_b);
public:
     JIT_PM(int k, double lambda, ReadableIndex &a, ReadableIndex &b);
     ~JIT_PM();
     virtual bool joinnext();
};

class Estimate {
public:
     int box_id_a, box_id_b;
     double score;
     Estimate() {}
     Estimate(int box_id_a, int box_id_b, double score);
     bool operator<(const Estimate& other) const;
};

Estimate::Estimate(int box_id_a, int box_id_b, double score)
     : box_id_a(box_id_a), box_id_b(box_id_b), score(score)
{
}

bool Estimate::operator<(const Estimate& other) const
{
     return score < other.score;
}

class Precompute_PM : public PM {
protected:
     Estimate *estimates;
public:
     Precompute_PM(int k, double lambda, ReadableIndex& a, ReadableIndex& b);
     ~Precompute_PM();
     virtual bool joinnext();
     int nabove(double threshold);
};

PM::PM(int k, double lambda, ReadableIndex &a, ReadableIndex &b)
     : a(a), b(b), score(lambda), results(k)
{
     neligible = a.nboxes() * b.nboxes();
}

JIT_PM::JIT_PM(int k, double lambda, ReadableIndex& a, ReadableIndex& b)
     : PM(k, lambda, a, b)
{
     eligibility = new std::vector<bool>(a.nboxes() * b.nboxes(), true);
}

JIT_PM::~JIT_PM()
{
}

inline
bool JIT_PM::eligible(int page_id_a, int page_id_b)
{
     return (*eligibility)[page_id_a * b.nboxes() + page_id_b];
}

void JIT_PM::mark_ineligible(int page_id_a, int page_id_b)
{ 
     (*eligibility)[page_id_a * b.nboxes() + page_id_b] = false;
     neligible--;
}

// Stupid version.
bool PM::joinnext()
{
     float s_max = FLT_MIN;
     int best_a, best_b;
     for (int i = 0; i < a.nboxes(); i++) {
          Box abox = a.get_box(i);
          for (int j = 0; j < b.nboxes(); j++) {
               join(i, j);
               return true;
          }
     }
     return false;
}

int iteration = 0;

bool JIT_PM::joinnext()
{
     float s_max = FLT_MIN;
     int best_a, best_b; // Box ID of the best pair seen.

     int c = 0;
     for (int i = 0; i < a.nboxes(); i++) {
          Box abox = a.get_box(i);
          for (int j = 0; j < b.nboxes(); j++) {
               if (!eligible(i, j)) continue;
               Box bbox = b.get_box(j);
               double ub = score.ub(abox, bbox);
               if (ub <= results.threshold) {
                    mark_ineligible(i, j);
                    continue;
               }
               double s = score.mean(abox, bbox);
               if (s <= s_max)
                    continue;
               s_max = s;
               best_a = i;
               best_b = j;
          }
     }
     if (iteration > 0 && iteration % 10 == 0) {
          printf("%d\n", iteration);
     }
     iteration++;
     if (s_max == FLT_MIN)
          return false;
#if 0
     printf("; joining %d and %d (%f--%f)\n", best_a, best_b, 
            score.lb(a.get_box(best_a), b.get_box(best_b)),
            score.ub(a.get_box(best_a), b.get_box(best_b)));
#endif
     join(best_a, best_b);
     return true;
}

void PM::join(int box_id_a, int box_id_b)
{
     const struct point *apoints, *bpoints;
     Box abox = a.get_box(box_id_a);
     Box bbox = b.get_box(box_id_b);
#ifndef NDEBUG
     double ub = score.ub(abox, bbox);
     double lb = score.lb(abox, bbox);
#endif
#if 1
     planesweep(abox, bbox, score.lambda, &results);
#else
#if 0
     nlj(abox, bbox, score.lambda, &results);
#else
     subboxes(abox, bbox, score.lambda, &results);
#endif
#endif
}

void JIT_PM::join(int box_id_a, int box_id_b)
{
     mark_ineligible(box_id_a, box_id_b);
     PM::join(box_id_a, box_id_b);
}

Precompute_PM::Precompute_PM(int k, double lambda, ReadableIndex& a,
                             ReadableIndex& b)
     : PM(k, lambda, a, b)
{
     int na = a.nboxes(), nb = b.nboxes();
     if (na * nb >= 2147483647 / sizeof(Estimate))
          fatal_error("Predication matrix too large for address space.");
     estimates = (Estimate *)mallock(na * nb * sizeof(Estimate));
     struct progress *p = make_progress(a.nboxes(), "Precomputing estimates");
     for (int i = 0; i < na; i++) {
          Box abox = a.get_box(i);
          for (int j = 0; j < nb; j++) {
               Box bbox = b.get_box(j);
               estimates[i * nb + j] = 
                    Estimate(i, j, score.ub(abox, bbox));
          }
          update_progress(p, i + 1);
     }
     free_progress(p);
     std::sort(estimates, estimates + na * nb);
}

Precompute_PM::~Precompute_PM()
{
//     delete[] estimates;
     free(estimates);
}

// Number of page pairs above a given threshold value.
int Precompute_PM::nabove(double threshold)
{
     int n = 0;
     int na = a.nboxes(), nb = b.nboxes();
     for (int i = 0; i < na; i++)
          for (int j = 0; j < nb; j++)
               if (estimates[i * nb + j].score >= threshold)
                    n++;
     return n;
}

#if 0
// mean
bool Precompute_PM::joinnext()
{
     if (neligible == 0)
          return false;
     neligible--;
     Estimate e = estimates[neligible];
     float ub = score.ub(a.get_box(e.box_id_a), b.get_box(e.box_id_b));
     iteration++;
     if (iteration % 100 == 0)
          printf("%d ub=%f\n", iteration, ub);
     if (results.threshold > ub)
          return true;
     join(e.box_id_a, e.box_id_b);
     return true;
}
#endif

// ub
bool Precompute_PM::joinnext()
{
     neligible--;
     Estimate e = estimates[neligible];
     iteration++;
     if (results.threshold > e.score)
          return false;
     if (iteration % 100 == 0)
          printf("%d: score %f >= threshold %f\n", iteration, e.score,
                 results.threshold);
     join(e.box_id_a, e.box_id_b);
     return neligible > 0;
}



static void usage(FILE *stream)
{
     fprintf(stream ,"Usage %s [-a FILENAME] [-b FILENAME] [-hv] FACTOR PAGE-SIZE\n", 
                  program_name);
     fprintf(stream, "Perform top-k probabilistic spatial join with prediction matrix.\n");
     if (stream == stderr)
          fprintf(stream, "Try `%s -h` for more information.\n", program_name);
}

static void help(void)
{
     usage(stdout);
     printf("\n"
"      -a FILENAME   First dataset.  (Renders FACTOR irrelevant.)\n"
"      -b FILENAME   Second dataset.  (Renders FACTOR irrelevant.)\n"
"      -h            Display this help.\n"
"      -v            Be verbose.\n"
          );
}

static bool verbose = false;
static int factor, page_size;
static std::string filename_a(""), filename_b("");

static void parse_options(int argc, char *argv[])
{
     int opt;

     while ((opt = getopt(argc, argv, "a:b:hv")) != -1)
          switch (opt) {
          case 'a': filename_a = optarg; break;
          case 'b': filename_b = optarg; break;
          case 'h': help(); exit(EX_OK);
          case 'v': verbose = true; break;
          default: usage(stderr); exit(EX_USAGE);
          }
     if (argc == optind + 2) {
          factor = atoi(argv[optind]);
          if (factor < 1)
               fatal_error("Factor must be at least one: %s", argv[optind]);
          page_size = atoi(argv[optind + 1]);
          if (page_size < 1)
               fatal_error("Page size must be at least one: %s",
                           argv[optind + 1]);
     } else {
          usage(stderr);
          exit(EX_USAGE);
     }
         
}

static std::string filename(const std::string base)
{
     std::stringstream s;
     s << "../experiments/" << factor << "x" << base << "-" << page_size;
     return s.str();
}


int main(int argc, char *argv[])
{
     set_program_name(argv[0]);
     parse_options(argc, argv);
     
     ReadableIndex a(filename_a == "" ? filename("7ca-33-x209y466") : filename_a);
     ReadableIndex b(filename_b == "" ? filename("7ca-33-x362y289") : filename_b);

     planesweep_verbose_on();
     mark_start_time(NULL);
     srandom(time(NULL));
     Precompute_PM pm(20, 1/223.0, a, b);
//     printf("Average volumes: %f and %f\n", a.average_volume(), b.average_volume());
//     printf("%d page pairs above %f\n", pm.nabove(0.003484), 0.003484);

     while (pm.joinnext())
          ;
     pm.results.print();
     printf("%s: %d %d %d %d %d %f\n", program_name, factor, page_size,
            a.npoints(), b.npoints(), iteration,
            elapsed_time(NULL));

     exit(0); // Just the return causes hang.  Don't know why.
     return EX_OK;
}

