#ifndef SCORE_H
#define SCORE_H

#include <math.h>

static inline double square(double x) { return x * x; }

#define MAX(A, B) ((A) >= (B) ? (A) : (B))
#define MIN(A, B) ((A) <= (B) ? (A) : (B))

class Score {
public:
     const double lambda;
     Score(double lambda) : lambda(lambda) {}
};

class Score_1D : public Score {
public:
     Score_1D(double lambda) : Score(lambda) {}
     double dist_ub(const Box box1, const Box box2) const {
          return MAX(box1.x1_max, box2.x1_max) - MIN(box1.x1_min, box2.x1_min);
     }
     double dist_lb(const Box box1, const Box box2) const {
          double d1 = 0;
          if (box2.x1_min - box1.x1_max > d1) d1 = box2.x1_min - box1.x1_max;
          if (box1.x1_min - box2.x1_max > d1) d1 = box1.x1_min - box2.x1_max;
          return d1;
     }
     double operator()(const Point a, const Point b) const {
          double d = fabs(a.x1 - b.x1);
          return a.p * b.p * lambda * exp(-lambda * d);
     }
     double ub(const Box box1, const Box box2) const {
          double d = dist_lb(box1, box2);
          double ub = box1.p_max * box2.p_max * lambda * exp(-lambda * d);
          assert(ub >= 0 && ub <= 1);
          return ub;
     }
     double lb(const Box box1, const Box box2) const {
          double d = dist_ub(box1, box2);
          double lb = box1.p_min * box2.p_min * lambda * exp(-lambda * d);
          assert(lb >= 0 && lb <= 1);
          return lb;
     }
     double mean(const Box box1, const Box box2) const {
          double x1 = (box1.x1_min + box1.x1_max) / 2;
          double x2 = (box2.x1_min + box2.x1_max) / 2;
          double d = fabs(x1 - x2);
          double p1 = (box1.p_min + box1.p_max) / 2;
          double p2 = (box2.p_min + box2.p_max) / 2;
          return p1 * p2 * lambda * exp(-lambda * d);
     }
};

class Score_L2 : public Score {
public:
     Score_L2(double lambda) : Score(lambda) {}
     double dist(const Point a, const Point b) const {
          return sqrt(square(a.x1 - b.x1) + square(a.x2 - b.x2));
     }
     double dist_ub(const Box box1, const Box box2) const {
          double d1 = 0, d2 = 0;
          if (box2.x1_min - box1.x1_max > d1) d1 = box2.x1_min - box1.x1_max;
          if (box1.x1_min - box2.x1_max > d1) d1 = box1.x1_min - box2.x1_max;
          if (box2.x2_min - box1.x2_max > d2) d2 = box2.x2_min - box1.x2_max;
          if (box1.x2_min - box2.x2_max > d2) d2 = box1.x2_min - box2.x2_max;
          return sqrt(square(d1) + square(d2));
     }
     double dist_lb(const Box box1, const Box box2) const {
          double d1 = MAX(box1.x1_max, box2.x1_max) - 
               MIN(box1.x1_min, box2.x1_min);
          double d2 = MAX(box1.x2_max, box2.x2_max) - 
               MIN(box1.x2_min, box2.x2_min);
          return sqrt(square(d1) + square(d2));
     }     
     double operator()(const Point a, const Point b) const {
          double pv = 1 - lambda * exp(-lambda * dist(a, b));
          return a.p * b.p * (1 - pv);
     }
     double ub(const Box box1, const Box box2) const {
          double d = dist_lb(box1, box2);
          double ub = box1.p_max * box2.p_max * lambda * exp(-lambda * d);
          assert(ub >= 0 && ub <= 1);
          return ub;
     }
     double lb(const Box box1, const Box box2) const {
          double d = dist_ub(box1, box2);
          double lb = box1.p_min * box2.p_min * lambda * exp(-lambda * d);
          assert(lb >= 0 && lb <= 1);
          return lb;
     }
     double mean(const Box box1, const Box box2) const {
          double x1 = (box1.x1_min + box1.x1_max) / 2;
          double y1 = (box1.x2_min + box1.x2_max) / 2;
          double x2 = (box2.x1_min + box2.x1_max) / 2;
          double y2 = (box2.x2_min + box2.x2_max) / 2;
          double d = sqrt(square(x1 - x2) + square(y1 - y2));
          double p1 = (box1.p_min + box1.p_max) / 2;
          double p2 = (box2.p_min + box2.p_max) / 2;
          return p1 * p2 * lambda * exp(-lambda * d);
     }
};

#endif
