/* sdist.cc: Compute distance and score distributions for a pair of boxes. */

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <vlcutils/mem.h>
#include <vlcutils/error.h>
#include <vlcutils/readline.h>
#include <vlcutils/progress.h>
#include <iostream>
#include <fstream>
#include "histogram.h"
#include "ddist.h"
#include "sdist.h"
#include "ptopk.h"

#define square(a) ((a) * (a))

#define NPART 4
#define NSAMPLES (NPART*NPART*NPART*NPART)

float *distsamples_for_boxes(Box a, Box b)
{
     float *samples = (float *)malloc(NSAMPLES * sizeof(float));
     if (!samples) fatal_perror("malloc");

     int count = 0;
     for (int i1 = 0; i1 < NPART; i1++) {
          double ax1 = a.x1_min + (i1 + 0.5) * 
               (a.x1_max - a.x1_min) / NPART;
          for (int i2 = 0; i2 < NPART; i2++) {
               double bx1 = b.x1_min + (i2 + 0.5) * 
                    (b.x1_max - b.x1_min) / NPART;
               double d1 = square(ax1 - bx1);
               for (int i3 = 0; i3 < NPART; i3++) {
                    double ax2 = a.x2_min + (i3 + 0.5) * 
                         (a.x2_max - a.x2_min) / NPART;
                    for (int i4 = 0; i4 < NPART; i4++) {
                         double bx2 = b.x2_min + (i4 + 0.5) * 
                              (b.x2_max - b.x2_min) / NPART;
                         double d2 = square(ax2 - bx2);
                         samples[count++] = sqrt(d1 + d2);
                    }
               }
          }
     }
     return samples;
}

static double sample_mean(float *samples, int nsamples)
{
     double sum = 0;
     for (int i = 0; i < nsamples; i++)
          sum += samples[i];
     return sum / nsamples;
}

static double sample_variance(float *samples, int nsamples, double mean)
{
     double sse = 0;
     for (int i = 0; i < nsamples; i++) {
          double e = (samples[i] - mean);
          sse += e * e;
     }
     return sse / (nsamples - 1);
}

struct distprofile *make_distprofile(Box box1, Box box2)
{
     float *samples = distsamples_for_boxes(box1, box2);
     struct distprofile *dp = 
          (struct distprofile *)mallock(sizeof(struct distprofile));
     dp->cdf = histogram_nbins(samples, NSAMPLES, 10);
#if 0
     fprintf(stderr, "Distances ranging from %f to %f.\n", dh->min, dh->max);
     plot_pdf(dh);
#endif
     return dp;
}

void free_distprofile(struct distprofile *dp)
{
     free(dp->cdf);
     free(dp);
}

static struct scoredist *scoredist_from_samples(float *samples, int nsamples,
                                                Box box1, Box box2)
{
     struct scoredist *sd = 
          (struct scoredist *)mallock(sizeof(struct scoredist));
     sd->cdf = histogram_nbins(samples, nsamples, 10);
     // Next, stretch the histogram based on the p-dimension of the boxes.
     sd->cdf->min *= box1.p_min * box2.p_min;
     sd->cdf->max *= box1.p_max * box2.p_max;
     free(samples);
     return sd;
}

struct scoredist *make_scoredist_exp(double lambda, Box box1, Box box2)
{
     float *samples = distsamples_for_boxes(box1, box2);
     for (int i = 0; i < NSAMPLES; i++)
          samples[i] = lambda * exp(-lambda * samples[i]);
     return scoredist_from_samples(samples, NSAMPLES, box1, box2);
}

struct scoredist *make_scoredist(DistDist *distdist, Box box1, Box box2)
{
     float *samples = distsamples_for_boxes(box1, box2);
#if 0
     struct histogram *dh = histogram_nbins(samples, NSAMPLES, 10);
     fprintf(stderr, "Distances ranging from %f to %f.\n", dh->min, dh->max);
     plot_pdf(dh);
     free(dh);
#endif
     for (int i = 0; i < NSAMPLES; i++)
          samples[i] = (1 - distdist->pvalue(samples[i]));
     return scoredist_from_samples(samples, NSAMPLES, box1, box2);
}

void free_scoredist(struct scoredist *sd)
{
     free(sd->cdf);
     free(sd);
}

float scoredist_pexceed(const struct scoredist *sd, float threshold)
{
     double pvalue = histogram_pvalue(sd->cdf, threshold);
     double pex = pow(pvalue, BOX_SIZE);
     printf("pexceed(%f) = %f\n", threshold, pex);
     return pex;
}

EVD::EVD(float k, float sigma, float mu)
     : k(k), sigma(sigma), mu(mu)
{
}

struct evd_param {
     double alpha, beta; /* Parameters for beta distribution. */
     EVD evd;
};

static struct evd_param *evdparams = NULL;
static int nevdparams = 0;

void load_evdparams()
{
     const char *filename = "evdparams.dat";
     FILE *f = fopen(filename, "r");
     if (!f) fatal_perror("%s", filename);
     nevdparams = fcount(f, '\n');
     fclose(f);
     evdparams = new struct evd_param[nevdparams];
     std::ifstream stream(filename, std::ios::in);
     for (int i = 0; i < nevdparams; i++) {
          stream >> evdparams[i].alpha;
          stream >> evdparams[i].beta;
          stream >> evdparams[i].evd.k;
          stream >> evdparams[i].evd.sigma;
          stream >> evdparams[i].evd.mu;
     }
     stream.close();
}

#define SQUARE(X) ((X) * (X))

void ScoreDistBeta::lookup_evdparams(double alpha, double beta)
{
     double min_d;
     int best_i;
     for (int i = 0; i < nevdparams; i++) {
          double d = SQUARE(alpha - evdparams[i].alpha) +
               SQUARE(beta - evdparams[i].beta);
          if (i == 0 || d < min_d) {
               min_d = d;
               best_i = i;
          }
     }
//     printf("Closest beta parameters: alpha = %f, beta = %f\n", 
//            evdparams[best_i].alpha, evdparams[best_i].beta);
     evd = evdparams[best_i].evd;
}

void normalize_sample(float *samples, double lb, double ub)
{
     for (int i = 0; i < NSAMPLES; i++)
          samples[i] = (samples[i] - lb) / (ub - lb);
}

ScoreDistBeta::ScoreDistBeta(DistDist *distdist, Box box1, Box box2)
{
     float *samples = distsamples_for_boxes(box1, box2);
     normalize_sample(samples, distdist->score_lb(box1, box2),
                      distdist->score_ub(box1, box2));
     double mean = sample_mean(samples, NSAMPLES);
     double variance = sample_variance(samples, NSAMPLES, mean);
//     free(samples);
     alpha = (1/mean - 1) * mean * mean * mean / variance - mean;
     beta = alpha / mean - alpha;
//     printf("Beta(%f, %f); mean = %f, variance = %f\n", alpha, beta, mean,
//            variance);
     if (!evdparams)
          load_evdparams();
     lookup_evdparams(alpha, beta);
}

ScoreDistBeta::~ScoreDistBeta()
{
     
}

float ScoreDistBeta::pexceed(float threshold)
{
     return evd.pexceed(threshold);
}

float EVD::pexceed(float threshold)
{
     if (k > -0.0001) k = -0.0001;
     double p = 1 - exp(-pow((1 + k * (threshold - mu) / sigma), -1/k));
//     printf("P(EVD(%f, %f, %f) > %f) = %f\n", k, sigma, mu, threshold, p);
     assert(0 <= p && p <= 1);
     return p;
}
