#include <sysexits.h>
#include <string>
#include <sstream>
#include <vlcutils/progress.h>
#include "ptopk.h"
#include "index.h"
#include "score.h"
#include "join.h" 

static void usage(FILE *stream)
{
     fprintf(stream ,"Usage %s [-a FILENAME] [-b FILENAME] [-hv] [nlj|ps|subboxes] FACTOR\n", 
                  program_name);
     fprintf(stream, "Perform probabilistic spatial join without prediction matrix.\n");
     if (stream == stderr)
          fprintf(stream, "Try `%s -h` for more information.\n", program_name);
}

static void help(void)
{
     usage(stdout);
     printf("\n"
"      -a FILENAME   First dataset.  (Renders FACTOR irrelevant.)\n"
"      -b FILENAME   Second dataset.  (Renders FACTOR irrelevant.)\n"
"      -h            Display this help.\n"
"      -v            Be verbose.\n"
          );
}

static bool verbose = false;
static enum { method_nlj, method_ps, method_subboxes } method;
static int factor, k = 20;
static std::string filename_a(""), filename_b("");

static void parse_options(int argc, char *argv[])
{
     int opt;

     while ((opt = getopt(argc, argv, "a:b:hv")) != -1)
          switch (opt) {
          case 'a': filename_a = optarg; break;
          case 'b': filename_b = optarg; break;
          case 'h': help(); exit(EX_OK);
          case 'v': verbose = true; break;
          default: usage(stderr); exit(EX_USAGE);
          }
     if (argc == optind + 2) {
          std::string method_string(argv[optind]);
          if (method_string == "nlj")
               method = method_nlj;
          else if (method_string == "ps")
               method = method_ps;
          else if (method_string == "subboxes")
               method = method_subboxes;
          else fatal_error("Unknown method: %s", method_string.c_str());
          factor = atoi(argv[optind + 1]);
          if (factor < 1)
               fatal_error("Factor must be at least one: %s",
                           argv[optind + 1]);
     } else {
          usage(stderr);
          exit(EX_USAGE);
     }
         
}

static std::string large1(const std::string base)
{
     std::stringstream s;
     s << "../experiments/large1-" << factor << "x" << base;
     return s.str();
}

int main(int argc, char *argv[])
{
     double lambda = 1 / 223.0;

     set_program_name(argv[0]);
     parse_options(argc, argv);

     std::string fna(filename_a == "" ? large1("7ca-33-x209y466") : filename_a);
     std::string fnb(filename_b == "" ? large1("7ca-33-x362y289") : filename_b);

     mark_start_time(NULL);
     ReadableIndex a(fna), b(fnb);
     if (verbose) {
          a.info();
          b.info();
     }

     // Cannot use object-level topk because large1 datasets don't
     // have object information.
     TopKResultsNoObjects reporter(k);
//     reporter.verbose = true;
     reporter.threshold = 1e-6;
     planesweep_verbose_on();
     
     Box abox = a.get_box(0);
     Box bbox = b.get_box(0);
     if (method == method_ps)
          planesweep(abox, bbox, lambda, &reporter);
     else if (method == method_nlj)
          nlj(abox, bbox, lambda, &reporter);
     else if (method == method_subboxes)
          subboxes(abox, bbox, lambda, &reporter);
     else fatal_error("Method not implemented.");

     reporter.print();
     printf("%d %d %f %d\n", abox.npoints, bbox.npoints, elapsed_time(NULL),
            reporter.iterations);
     return 0;
}
